#include "yacc2html.h"

/*  yacc2html -- a YACC-input-to-HTML converter.
 *
 *  Copyright 1994 Jutta Degener, jutta@cs.tu-berlin.de, Technical
 *  University of Berlin.  No warranty implied, all rights reserved.
 */

char const * program_name;	/* argv[0], used for error messages 	*/
char const * input_file;	/* argv[argc - 1], or *standard input*	*/

/* Basename, set with -b option.  Used for %b format string in
 * various file and link names.
 * Default is the input file name without file appendix, if any.
 */
char const * basis;

/* The name of a file, specified with -Y if any, to which a sed script
 * is written which augments names of nonterminals with links to the
 * yacc html file.
 */
char const * yacc_sed_file;

/* The name of the yacc html file, specified with -y if any, that
 * is used for the references in the lex and yacc sed scripts.
 * Default is "%b-y.html".  (%b is the basename of the input file,
 * or whatever the user specified with -b on the command line.)
 */
char const * yacc_file;

/* The name of a file, specified with -L if any, to which a sed
 * script is written which turns a lex file into an html file with
 * <a name="..."> anchors around the first occurance of each token.
 */
char const * lex_sed_file;

/* The name of the lex html file, specified with -l if any, that
 * is assumed for the references in the output.   Default is "%b-l.html".
 */
char const * lex_file;

/* -r flag: turn nonterminals into hrefs, even if they refer to the
 * 	    current definition. (Default: don't.)
 */
int recursive_links;

/* -u flag: link type tags to the %union statement, if any. 
 *          (Default: don't.)
 */
int union_links;

/* -t flag: link terminals to their %token definition, rather than to a
 *	    lex file.  (Default: don't.)
 */
int token_terminals;

/* -i flag: don't link terminals anywhere.  (Default: link them to their
 * 	    lex file.)
 */
int ignore_terminals;

/* -c flag: do not output the actions in the grammar, just the rules.
 * (YACC's C-style comments are not considered C code in this sense,
 * they're white space.)  (Default: do print them.)
 */
int c_code_suppression;

/* document <title>, set with -h option.
 */
char const * title;	

/* format string for generated URLs of terminals (lex tokens); set with -T
 * option.  (Default: if terminals are local, "%s", otherwise, "%l#%s",
 * where %l is the name of the lex file.)
 */
char const * fmt_terminal_url;

/* format string for generated URLs of nonterminals; set with -N option.
 * Use -N to enforce naming conventions with your nonterminals, e.g.
 * -N "yacc-%s" will prefix nonterminal URLs with "yacc-".
 */
char const * fmt_nonterminal_name = "%s";

static void usage(void)
{
	fprintf(stderr, "\
Usage: %s [-b basename] [-c] [-D name[=url]] [-h title] [-H] [-i]\n\
	[-l lex-file] [-L lex-sed-file] [-N nonterminal-format]\n\
	[-p] [-r] [-t] [-T terminal-format] [-u] [-y yacc-file]\n\
	[-Y yacc-sed-file] [filename]\n", program_name);

	exit(64);
}


static void help(void)
{
	fprintf(stderr,
"%s -- convert a yacc grammar to HTML.\n\n", program_name);

	fprintf(stderr,
"\"%s grammar.y\" writes a HTML version of grammar.y to standard output.\n\
For a detailed description of the options, see the manual page.\n\
\n\
Options are:\n", program_name);
	fputs("\
   -b basis   	set basename to <basis>\n\
   -l file	consider <file> the name of the lex HTML file\n\
   -y file 	consider <file> the name of the yacc HTML file\n\
   -h header  	use <header> as the title in the output\n\
\n", stderr);
	fputs("\
   -Dname     	do not link <name> to anything\n\
   -Dname=url   link <name> to URL <url>\n\
   -t		link terminals to their %token definition, not to a lex file\n\
   -i		ignore terminals, do not link them\n\
   -r 		link in recursive nonterminal from their own definition\n\
   -u 		link type tags (in angle brackets) to the %union definition\n\
\n", stderr);
	fputs("\
   -N fmt	generate nonterminal names from <fmt> (%s=the name)\n\
   -T fmt       generate terminal URLs from <fmt> (%s=the name)\n\
\n\
   -L file	write a sed script to create lex HTML to <file>\n\
   -Y file      write a sed script to anchor nonterminals to <file>\n\
   -c 	      	strip c code from HTML output\n\
   -H	      	print this text\n\n",
   	stderr);

	exit(0);
}

int main(int ac, char ** av)
{
	extern int	  optind;
	extern char	* optarg;

	int		  opt;
	char		* c;
	FILE 		* f;

	/* malloc_debug(2); */

	program_name = av[0] ? av[0] : "c";
	if ((c = strrchr(program_name, '/')) && c[1]) program_name = c + 1;

	while ((opt=getopt(ac,av,"b:cdD:Hh:il:L:N:rtT:uy:Y:")) != EOF)
	switch (opt) {
	case 'H':	help(); /* notreached */
#ifndef	NDEBUG
	/* case 'd':	malloc_debug(2);		break; */
#endif
	case 'c':	c_code_suppression   = 1; 	break;
	case 'r':	recursive_links      = 1; 	break;
	case 'u':	union_links 	     = 1;	break;
	case 'h':	title 		     = optarg; 	break;

	case 'T':	fmt_terminal_url     = optarg; 	break;
	case 'N':	fmt_nonterminal_name = optarg; 	break;

	case 'L':	lex_sed_file 	     = optarg; 	break;
	case 'l':	lex_file 	     = optarg; 	break;

	case 'Y':	yacc_sed_file 	     = optarg; 	break;
	case 'y':	yacc_file 	     = optarg; 	break;

	case 'b':	basis 		     = optarg;	break;

	case 'i':	ignore_terminals = 1;
			if (token_terminals) goto both_terminals;
			break;

	case 't':	token_terminals = 1;
			if (ignore_terminals) {
both_terminals:			fprintf(stderr,
	"%s: use -i (ignore)  or -t (token) terminals, but not both.\n",
					program_name);
				usage();
			}
			break;

	case 'D':	if (!(c = strchr(optarg, '=')))
				define_link(optarg, (char *)0);
			else {
				char * p = TMALLOC(char, 1 + c - optarg);
				memcpy(p, optarg, c - optarg);
				p[ c - optarg ] = 0;
				define_link(p, c + 1);
				VFREE(p);
			}
			break;
		
	default:  	usage();
	}

	ac -= optind;
	av += optind;

	if (!fmt_terminal_url) {
		if (token_terminals) fmt_terminal_url = "%s";
		else if (!ignore_terminals) fmt_terminal_url = "%l#%s";
	} else if (ignore_terminals) {
		fprintf(stderr, 
	"%s: specify either -i or a format for terminal URLs, but no both.\n",
		program_name);
		usage();
	}

	if (ac > 1) usage();

	if (ac <= 0) {
		f = stdin;
		if (!basis) basis = "stdin";
		input_file = "*standard input*";
	} else {
		if (!(f = fopen(*av, "r"))) {
			perror(*av);
			exit(1);
		}
		if (!basis) {
			char * p, * s = TMALLOC(char, strlen(*av) + 1);
			strcpy(s, *av);
			if ((p = strrchr(s, '.')) && !strchr(p, '/')) *p = 0;
			basis = intern(s); 
			VFREE(s);
		}
		input_file = *av;
	}

	if (isatty(fileno(f)) && isatty(fileno(stdout)))
		parser_set_interactive(f);
	else parser_set_file(*av, f);

	yacc_file = intern(uprint(yacc_file ? yacc_file : "%b-y.html", "%s"));
	lex_file  = intern(uprint(lex_file  ? lex_file  : "%b-l.html", "%s"));

	yyparse();
	terminate_output();

	if (parser_print_errors()) exit(1);

	if (lex_sed_file && write_lex_file( lex_sed_file)) exit(1);
	if (yacc_sed_file && write_yacc_file( yacc_sed_file)) exit(1);

	exit(0);
}
